@@ -1,14 +1,17 @@
 import torch
-from pytorch_lightning import LightningModule
+import pytorch_lightning as pl
+from pytorch_lightning.accelerators import CPUAccelerator
 from pytorch_lightning.plugins import HorovodPlugin
 from pytorch_lightning.utilities import rank_zero_only
 
 import ray
 from ray import ObjectRef
+from ray.util import PublicAPI
 from ray.util.queue import Queue
 
 from ray_lightning.session import init_session
-from ray_lightning.util import process_results, Unavailable
+from ray_lightning.util import process_results, Unavailable, to_state_stream, \
+    load_state_stream
 from ray_lightning.tune import TUNE_INSTALLED, is_session_enabled
 
 try:
@@ -28,6 +31,7 @@ def get_executable_cls():
     return None
 
 
+@PublicAPI(stability="beta")
 class HorovodRayPlugin(HorovodPlugin):
     """Pytorch Lightning Plugin for Horovod training on a Ray cluster.
 
@@ -45,8 +49,8 @@ class HorovodRayPlugin(HorovodPlugin):
     script: ``python train.py``, and not with ``horovodrun``.
 
     Args:
-        num_hosts (int): The number of nodes/machines to execute the job on.
-        num_slots (int): Number of workers to be placed on each machine.
+        num_workers (int): Number of training workers to use.
+        num_cpus_per_worker (int): Number of CPUs per worker.
         use_gpu (bool): Whether to use GPU for allocation. For GPU to be
             used, you must also set the ``gpus`` arg in your Pytorch Lightning
             Trainer to a value > 0.
@@ -59,20 +63,18 @@ class HorovodRayPlugin(HorovodPlugin):
             from ray_lightning import HorovodRayPlugin
 
             ptl_model = MNISTClassifier(...)
-            # 2 nodes, 4 workers per node, each using 1 CPU and 1 GPU.
-            plugin = HorovodRayPlugin(num_hosts=2, num_slots=4,
-                use_gpu=True)
+            plugin = HorovodRayPlugin(num_workers=2, use_gpu=True)
 
-            # If using GPUs, set the ``gpus`` arg to a value > 0.
-            # The actual number of GPUs is determined by ``num_slots``.
-            trainer = pl.Trainer(..., gpus=1, plugins=[plugin])
+            # Don't set ``gpus`` in ``Trainer``.
+            # The actual number of GPUs is determined by ``num_workers``.
+            trainer = pl.Trainer(..., plugins=[plugin])
             trainer.fit(ptl_model)
 
     """
 
     def __init__(self,
-                 num_hosts: int = 1,
-                 num_slots: int = 1,
+                 num_workers: int,
+                 num_cpus_per_worker: int = 1,
                  use_gpu: bool = False):
 
         if not HOROVOD_AVAILABLE:
@@ -81,8 +83,8 @@ class HorovodRayPlugin(HorovodPlugin):
             ray.init()
         super().__init__()
         self.nickname = "horovod_ray"
-        self.num_hosts = num_hosts
-        self.num_slots = num_slots
+        self.num_workers = num_workers
+        self.cpus_per_worker = num_cpus_per_worker
         self.use_gpu = use_gpu
         self.executor = None
 
@@ -110,20 +112,37 @@ class HorovodRayPlugin(HorovodPlugin):
     @property
     def world_size(self) -> int:
         if not hvd.is_initialized():
-            return self.num_hosts * self.num_slots
+            return self.num_workers
         return hvd.size()
 
-    def setup(self, model: LightningModule):
+    def setup(self):
         """Creates the RayExecutor object."""
-        self._model = model
         settings = RayExecutor.create_settings(timeout_s=30)
         self.executor = RayExecutor(
             settings,
-            num_hosts=self.num_hosts,
-            num_slots=self.num_slots,
+            num_workers=self.num_workers,
+            cpus_per_worker=self.cpus_per_worker,
             use_gpu=self.use_gpu)
         self.executor.start(executable_cls=get_executable_cls())
 
+    def setup_environment(self) -> None:
+        # Swap out the accelerator if necessary.
+        # This is needed to support CPU head with GPU workers or Ray Client.
+        current_accelerator = self.lightning_module.trainer.accelerator
+        if self.use_gpu and isinstance(current_accelerator, CPUAccelerator):
+            from weakref import proxy
+            from ray_lightning.util import DelayedGPUAccelerator
+            precision_plugin = current_accelerator.precision_plugin
+            new_accelerator = DelayedGPUAccelerator(
+                precision_plugin=precision_plugin, training_type_plugin=self)
+            self.lightning_module.trainer._accelerator_connector \
+                ._training_type_plugin = \
+                proxy(new_accelerator.training_type_plugin)
+            self.lightning_module.trainer._accelerator_connector \
+                ._precision_plugin = proxy(new_accelerator.precision_plugin)
+            self.lightning_module.trainer._accelerator_connector.accelerator \
+                = new_accelerator
+
     def pre_dispatch(self):
         """All pre-dispatch logic should be done in train_remote instead."""
         pass
@@ -150,7 +169,8 @@ class HorovodRayPlugin(HorovodPlugin):
 
         results = process_results(result_futures, queue)
 
-        results, state_dict, best_path = results[0]
+        results, state_stream, best_path = results[0]
+        state_dict = load_state_stream(state_stream, to_gpu=self.use_gpu)
         self._results = results
         self._model = model
         self._model.load_state_dict(state_dict)
@@ -168,9 +188,10 @@ class HorovodRayPlugin(HorovodPlugin):
     def train_remote(self, model: ObjectRef, queue: Queue = None, **kwargs):
         """Training function to be executed on each remote worker."""
         self._model = ray.get(model)
-        self.lightning_module.trainer.accelerator_connector\
+        self.lightning_module.trainer._accelerator_connector\
             ._training_type_plugin = self
-        self.lightning_module.trainer.accelerator.training_type_plugin = self
+        self.lightning_module.trainer._accelerator_connector.accelerator\
+            .training_type_plugin = self
 
         hvd.init()
         rank_zero_only.rank = self.global_rank
@@ -196,9 +217,10 @@ class HorovodRayPlugin(HorovodPlugin):
                 self.lightning_module.trainer.checkpoint_callback\
                     .best_model_path
 
-        return results, self.lightning_module.state_dict(), best_model_path
+        return results, to_state_stream(self.lightning_module.state_dict()), \
+            best_model_path
 
-    def post_dispatch(self):
+    def post_dispatch(self, trainer: "pl.Trainer"):
         """Shuts down the RayExecutor."""
         self.executor.shutdown()
 
@@ -209,6 +231,9 @@ class HorovodRayPlugin(HorovodPlugin):
     @property
     def root_device(self):
         if self.use_gpu and torch.cuda.is_available():
-            return torch.device("cuda", hvd.local_rank())
+            if hvd.is_initialized():
+                return torch.device("cuda", hvd.local_rank())
+            else:
+                return torch.device("cuda", 0)
         else:
             return torch.device("cpu")
